"""Base class for integration tests, based on ZopeTestCase and PloneTestCase.

Note that importing this module has various side-effects: it registers a set of
products with Zope, and it sets up a sandbox Plone site with the appropriate
products installed.
"""

from Testing import ZopeTestCase

# Let Zope know about the products we require above-and-beyond a basic
# Plone install (PloneTestCase takes care of these).
ZopeTestCase.installProduct('ArchAddOn')

# Import PloneTestCase - this registers more products with Zope as a side effect
from Products.PloneTestCase.PloneTestCase import PloneTestCase
from Products.PloneTestCase.PloneTestCase import FunctionalTestCase
from Products.PloneTestCase.PloneTestCase import setupPloneSite

# Set up a Plone site, and install the ArchAddOn product.
setupPloneSite(products=('ArchAddOn',))

class ArchAddOnTestCase(PloneTestCase):
    """Base class for integration tests for the 'ArchAddOn' product. This may
    provide specific set-up and tear-down operations, or provide convenience
    methods.
    """

class ArchAddOnFunctionalTestCase(FunctionalTestCase):
    """Base class for functional integration tests for the 'ArchAddOn' product.
    This may provide specific set-up and tear-down operations, or provide
    convenience methods.
    """




from base import ArchAddOnTestCase
from Products.Archetypes.Extensions.utils import installTypes
from Products.Archetypes.public import listTypes
from Products.ArchAddOn.config import PROJECTNAME
from StringIO import StringIO

class TestFields(ArchAddOnTestCase):

    def afterSetUp(self):
        out=StringIO()
        installTypes(self.portal, out,
                     listTypes(PROJECTNAME),
                     PROJECTNAME)
        self.types = ('ArchAddOnExample',)
        self.setRoles(('Manager',))
        self.portal.invokeFactory('ArchAddOnExample', 'aao')
        self.aao = self.portal.aao

    # Test these fields
    #    USAddressField( 'usaddress' ),
    #    LinkField( 'link' ),
    #    USPhoneField( 'usphone' ),
    #    EmailField( 'email' ),
    #    InstructionField( 'instruction', widget=InstructionWidget( description='This is an instruction' )),
    #    DynamicField( 'dynamic' ),
    #    SimpleDataGridField('datagrid'),

    def testSimpleDataGridField(self):
        self.aao.setDatagrid('row1|Row 1|Description of Row 1')
        self.assertEqual(self.aao.getField('datagrid').lookup(self.aao, 'row1', 1), 'Row 1')

    # XXX: Add tests for the rest of the fields.

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestFields))
    return suite



from Products.Archetypes.public import listTypes
from Products.Archetypes.Extensions.utils import installTypes, install_subskin
from Products.ArchAddOn.config import PROJECTNAME, GLOBALS
from Products.ArchAddOn.config import INSTALL_SAMPLE_TYPES

from StringIO import StringIO


def install(self):
    out = StringIO()

    if INSTALL_SAMPLE_TYPES:
        installTypes(self, out,
                     listTypes(PROJECTNAME),
                     PROJECTNAME)

    install_subskin(self, out, GLOBALS)

    print >> out, "Successfully installed %s." % PROJECTNAME
    return out.getvalue()



from Products.Archetypes import atapi
from Products.Archetypes.public import DisplayList
from Widgets import LinkWidget, USPhoneWidget, EmailWidget, AddressWidget, \
    InstructionWidget, SimpleDataGridWidget


class USAddressField(atapi.TextField):
    """A field that stores strings"""
    _properties = atapi.TextField._properties.copy()
    _properties.update(
        {'widget': AddressWidget,
         'validators': ('isUSAddress',),
         })


class LinkField(atapi.StringField):
    """A field that stores strings"""
    _properties = atapi.StringField._properties.copy()
    _properties.update(
        {'widget': LinkWidget,
         'validators': ('isURL',),
         })


class USPhoneField(atapi.StringField):
    """A field that stores strings"""
    _properties = atapi.StringField._properties.copy()
    _properties.update(
        {'widget': USPhoneWidget,
         'validators': ('isFormattedUSPhone',),
         })


class EmailField(atapi.StringField):
    """A field that stores strings"""
    _properties = atapi.StringField._properties.copy()
    _properties.update(
        {'widget': EmailWidget,
         'validators': ('isEmail',),
         })


class InstructionField(atapi.ObjectField):
    """Just help"""
    _properties = atapi.ObjectField._properties.copy()
    _properties.update({'widget': InstructionWidget})


class SimpleDataGridField(atapi.LinesField):
    """A lines field with embedded vertical bars for fields. If the
    columns property is set, the isValidGrid validator will ensure that
    the entered text has exactly that number of columns for each row.
    If strip_whitespace is True, the individual cells in the grid will
    be stripped of leading and trailing whitespace (that is, whitespace
    added around the delimiter).

    If column_names is set to a tuple or list of strings, it gives the names
    of columns in the grid. The getDataGridRow script and the getRow() method
    will return a dict using these column names as keys rather than using the
    column numbers (starting from zero).

    Fields are just strings, with no internal validation, etc.
    """
    _properties = atapi.LinesField._properties.copy()
    _properties.update(
        {'widget': SimpleDataGridWidget,
         'columns': None,
         'column_names': None,
         'validators': ('isValidGrid',),
         'strip_whitespace': True,
         'delimiter': '|',
         })

    def get(self, instance, **kwargs):
        data = atapi.LinesField.get(self, instance, **kwargs)
        return data

    def getAsGrid(self, instance, **kwargs):
        """Return a tuple of tuples - the outer tuple has one element
        per row in the grid, the inner tuple has one element per column
        in that row.
        """
        data = self.get(instance, **kwargs)
        rows = []
        for d in data:
            cols = self._split(d)
            rows.append(tuple(cols))
        return tuple(rows)

    def getRow(self, instance, key, keyCol=0, **kwargs):
        """Get a row matching the given key, looked up in the given column.

        Returns a dict. If the column_names property is not set, the
        keys of the dict are the integral column indexes. If it is
        set, the relevant column names are used as keys instead.
        """
        data = self.get(instance, **kwargs)
        for d in data:
            rowData = self._split(d)
            if rowData[keyCol] == self._strip(key):
                row = {}
                for i in range(0, len(rowData)):
                    key = self._getColName(i)
                    row[key] = rowData[i]
                return row

    def getColumn(self, instance, column, **kwargs):
        """Get a tuple of all values in the given column, indexed from 0.
        """
        data = self.get(instance, **kwargs)
        col = []
        for d in data:
            col.append(self._split(d)[column])
        return tuple(col)

    def lookup(self, instance, key, column, keyCol=0, **kwargs):
        """Look for the given key in the column with index given by keyCol,
        and return the value stored in the given column. Returns None if the
        value could not be found. Raises IndexError if the matching row
        does not have the requested column. Returns None if the value could
        not be found.
        """
        data = self.get(instance, **kwargs)
        for d in data:
            cols = self._split(d)
            if cols[keyCol] == self._strip(key):
                return cols[column]
        return None

    def getAsDisplayList(self, instance, keyCol=0, valueCol=1, **kwargs):
        """Get two columns of each row as a DisplayList - the key columns is
        keyCol, and the value column is valueCol.
        """
        data = self.get(instance, **kwargs)
        lst = DisplayList()
        for d in data:
            cols = self._split(d)
            lst.add(cols[keyCol], cols[valueCol])
        return lst

    def _strip(self, value):
        """Strip whitespace from a value if applicable"""
        if self.strip_whitespace:
            return value.strip()
        else:
            return value

    def _split(self, row):
        """Return a row split on the delimiter, optionally stripped
        """
        return [self._strip(r) for r in row.split(self.delimiter)]

    def _getColName(self, column):
        """Get the name of the given column. If column_names is set, use
        that if it contains a value. Else, use the integral column name.
        """
        if self.column_names and len(self.column_names) > column:
            return self.column_names[column]
        else:
            return column

#-------- dyndoc helper classes

from AccessControl import getSecurityManager
from zope.tal.taldefs import TALExpressionError as TALESError

from Products.PageTemplates import PageTemplate
from Products.PageTemplates.ZopePageTemplate import SecureModuleImporter
from Acquisition import Implicit, aq_parent, aq_inner
from Persistence import Persistent


class ddPageTemplate(PageTemplate.PageTemplate, Implicit, Persistent):

    def pt_getContext(self):
        root = self.getPhysicalRoot()
        template = aq_parent(self)
        here = aq_parent(template)
        c = {'template': template,
             'options': {},
             'nothing': None,
             'request': None,
             'root': root,
             'request': getattr(root, 'REQUEST', None),
             'modules': SecureModuleImporter,
             'here': here,
             'context': here,
             'container': aq_inner(here),
             }
        return c

    def __setstate__(self, state):
        ddPageTemplate.inheritedAttribute('__setstate__')(self, state)
        self._cook()

#class Macros(Base):
#
# taken from CMFDynamicDocument
#
# should be easy to allow dyndocs to be used as a macro (uncomment
# this and stuff below), but not tested/not interested in this.

#    """
#    Evaluate to the macros collection of the contained page template.
#    """
#    def __of__( self, doc):
#        return doc.get_pt().macros


class DynamicField(atapi.TextField):
    """Just help"""
    _properties = atapi.ObjectField._properties.copy()
    _properties.update(
        {'widget': atapi.TextAreaWidget,
         'default_output_type': 'text/html',
         })

    #macros = Macros()
    #security.declareProtected(CMFCorePermissions.View, 'macros')

    def get_unwrapped(self, instance):
        """
        return unwrapped PageTemplate object
        """

        #import pdb
        #pdb.set_trace()
        #raise "foo",dir(fldname)
        try:
            return getattr(instance, self.getName()).__of__(instance)
        except:
            setattr(instance, self.getName(), ddPageTemplate())
            return getattr(instance, self.getName()).__of__(instance)

    def getRaw(self, instance, **kwargs):
        """Get raw"""

        # not sure if there's a better way to do this:
        # if this is our first access (ie, we haven't used property
        # before, then init to page template)
        # probably better handled in object's __init__ or addXXXX function,
        # both not sure how to handle __init__ properly, and
        # addXXXX function would make user of this fieldtype have to
        # create one for every class that used this field type.

        if not hasattr(instance, self.getName()):
            setattr(instance, self.getName(), ddPageTemplate())
        else:
            return self.get_unwrapped(instance).read()

    def set(self, instance, value, **kwargs):
        """
          Edit the document
        """
        pt = self.get_unwrapped(instance)
        pt.write(value)

    def get(self, instance, mimetype=None, raw=0, **kwargs):
        """Return"""
        pt = self.get_unwrapped(instance)

        if raw:
            # when would this happen? not sure, copying from TextField
            return pt.read()

        security = getSecurityManager()
        security.addContext(self)
        try:
            try:
                result = self.get_unwrapped(instance).pt_render(
                    extra_context={'user': security.getUser()})
            except TALESError, err:
                if err.type == 'Unauthorized':
                    raise err.type, err.value, err.traceback
                raise
            return result
        finally:
            security.removeContext(self)


__all__ = ('USAddressField', 'LinkField', 'USPhoneField', 'EmailField',
           'InstructionField', 'DynamicField', 'SimpleDataGridField')



from zope.interface import implements
from Products.validation.interfaces.IValidator import IValidator
import re

phone_pattern = re.compile('\(\d\d\d\) \d\d\d-\d\d\d\d')


class FormattedUSPhoneValidator:
    implements(IValidator)

    def __init__(self, name):
        self.name = name
        return None

    def __call__(self, value, *args, **kwargs):
        if not phone_pattern.match(value):
            return """Not in form: (999) 555-1212"""
        return 1


class USAddressValidator:
    implements(IValidator)

    def __init__(self, name):
        self.name = name
        return None

    def __call__(self, value, *args, **kwargs):
        if not ("\n" in value or "\r" in value):
            return """Not a valid US addresss."""
        return 1


class LinesAllFloatValidator:
    implements(IValidator)

    def __init__(self, name):
        self.name = name
        return None

    def __call__(self, value, *args, **kwargs):
        for i in value:
            try:
                float(i)
            except:
                return """Not a valid floating-point number: %s""" % i
        return 1


class LinesAllIntValidator:
    implements(IValidator)

    def __init__(self, name):
        self.name = name
        return None

    def __call__(self, value, *args, **kwargs):
        for i in value:
            try:
                int(i)
            except:
                return """Not a valid integer: %s""" % i
        return 1


class SimpleDataGridValidator:
    """Ensure a data grid field has the correct number of columns, if a value
    is set as a field property.
    """
    implements(IValidator)

    def __init__(self, name):
        self.name = name
        return None

    def __call__(self, value, *args, **kwargs):
        field = kwargs['field']
        cols = field.columns
        if cols:
            for v in value:
                if len(v.split(field.delimiter)) != cols:
                    return "Each row must have exactly %d fields" % cols
        return 1


__all__ = ('FormattedUSPhoneValidator', 'USAddressValidator',
           'LinesAllFloatValidator', 'LinesAllIntValidator',
           'SimpleDataGridValidator')



from base import ArchAddOnTestCase
from Products.Archetypes.Extensions.utils import installTypes
from Products.Archetypes.public import listTypes
from Products.ArchAddOn.config import PROJECTNAME
from StringIO import StringIO

class TestProductInstall(ArchAddOnTestCase):

    def afterSetUp(self):
        out=StringIO()
        installTypes(self.portal, out,
                     listTypes(PROJECTNAME),
                     PROJECTNAME)
        self.types = ('ArchAddOnExample',)

    def testTypesInstalled(self):
        for t in self.types:
            self.failUnless(t in self.portal.portal_types.objectIds(),
                            '%s content type not installed' % t)

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestProductInstall))
    return suite



from Products.Archetypes.Widget import StringWidget, TextAreaWidget
from Products.Archetypes.Widget import LinesWidget


class LinkWidget(StringWidget):
    _properties = StringWidget._properties.copy()
    _properties.update(
        {'size': '60',
         'maxlength': '255',
         'description': ("Enter the web address (URL). You can copy & paste "
                         "this from a browser window."),
         'macro': 'archaddon_widgets/link',
         })


class USPhoneWidget(StringWidget):
    _properties = StringWidget._properties.copy()
    _properties.update(
        {'size': '20',
         'maxlength': '20',
         'description': ("Enter phone number in the form '(800) 555-1212'. "
                         "Don't forget the area code."),
         })


class EmailWidget(StringWidget):
    _properties = StringWidget._properties.copy()
    _properties.update(
        {'size': '40',
         'maxlength': '60',
         'macro': 'archaddon_widgets/email',
         })


class AddressWidget(TextAreaWidget):
    _properties = TextAreaWidget._properties.copy()
    _properties.update(
        {'rows': 3,
         'cols': 40,
         'description': ("Enter the address, putting each address line on a "
                         "separate line."),
         #'macro':'archaddon_widgets/address'
         })


class InstructionWidget(StringWidget):
    _properties = StringWidget._properties.copy()
    _properties.update(
        {'macro': 'archaddon_widgets/instruction',
         'modes': ('edit',),
         'visible': {'view': 0, 'edit': 'visible'},
         })


class SimpleDataGridWidget(LinesWidget):
    _properties = LinesWidget._properties.copy()
    #_properties.update(
    #    {#'macro':'archaddon_widgets/instruction',
    #        })


__all__ = ('LinkWidget', 'USPhoneWidget', 'EmailWidget', 'AddressWidget',
           'InstructionWidget', 'SimpleDataGridWidget')



from setuptools import setup, find_packages
import os

versionfile = open(os.path.join('Products', 'ArchAddOn', 'version.txt'))
version = versionfile.read().strip()
versionfile.close()

readmefile = open(os.path.join('Products', 'ArchAddOn', 'README.txt'))
readme = readmefile.read().strip()
readmefile.close()

historyfile = open(os.path.join('docs', 'HISTORY.txt'))
history = historyfile.read().strip()
historyfile.close()

long_description = readme + "\n\n" + history

setup(name='Products.ArchAddOn',
      version=version,
      description="Straightforward toolbox of field types, widgets, and validators for Archetypes.",
      long_description=long_description,
      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='',
      author='Joel Burton',
      author_email='joel@joelburton.com',
      url='https://svn.plone.org/svn/collective/Products.ArchAddOn',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'Plone>=4.0dev',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



## Script (Python) "getDataGridRow"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=fieldName, key, keyColumn=0
##title=Get the entire data-grid row of a given field

# Looks for the given data grid field in the current context and its
# acquisition parents. Once found, look up the given key in the given column
# (defaults to 0 - the first column) and return the row.

# Returns a dict. If the field has the column_names property set, use these
# as the dict keys. Else use the column index as keys, starting from 0.

obj = context

while obj is not None:
    if hasattr(obj, 'Schema') and obj.Schema().has_key(fieldName):
        # Found it
        break
    else:
        if hasattr(obj, 'aq_parent'):
            obj = obj.aq_parent
        else:
            obj = None
            break

if obj is None:
    raise AttributeError, "Field %s not found" % (fieldName,)

return obj.getWrappedField(fieldName).getRow(obj, key, keyColumn)



from Products.Archetypes.public import process_types, listTypes
from Products.CMFCore import utils
from Products.CMFCore.DirectoryView import registerDirectory

from Products.validation import validation
from Validators import FormattedUSPhoneValidator, USAddressValidator, \
    LinesAllFloatValidator, LinesAllIntValidator, SimpleDataGridValidator
validation.register(FormattedUSPhoneValidator('isFormattedUSPhone'))
validation.register(USAddressValidator('isUSAddress'))
validation.register(LinesAllFloatValidator('areLinesFloat'))
validation.register(LinesAllIntValidator('areLinesInt'))
validation.register(SimpleDataGridValidator('isValidGrid'))

from config import ADD_CONTENT_PERMISSION, PROJECTNAME, SKINS_DIR, GLOBALS

registerDirectory(SKINS_DIR, GLOBALS)


def initialize(context):
    """Initializer called when used as a Zope 2 product."""

    #Import Types here to register them
    import Example

    content_types, constructors, ftis = process_types(
        listTypes(PROJECTNAME),
        PROJECTNAME)

    utils.ContentInit(
        PROJECTNAME + ' Content',
        content_types=content_types,
        permission=ADD_CONTENT_PERMISSION,
        extra_constructors=constructors,
        fti=ftis,
        ).initialize(context)



from Products.Archetypes import atapi
from Fields import USAddressField, LinkField, USPhoneField, EmailField, \
    InstructionField, SimpleDataGridField
# Untested due to problems:
# from Fields import DynamicField
from Products.ArchAddOn.Widgets import InstructionWidget, SimpleDataGridWidget
from Products.ArchAddOn.config import PROJECTNAME


Schema = atapi.BaseSchema.copy() + atapi.Schema((
    USAddressField('usaddress'),
    LinkField('link'),
    USPhoneField('usphone'),
    EmailField('email'),
    InstructionField(
        'instruction',
        widget=InstructionWidget(description='This is an instruction')),

    #XXX: This breaks the example so better to fix or not include.
    #DynamicField( 'dynamic' ),

    SimpleDataGridField('datagrid',
        column_names=('col1', 'col2', 'col3'),
        columns=3,
        default=[
            'row1|Row 1|Row 1 Description',
            'row2|Row 2|Row 2 Description',
        ],
        widget=SimpleDataGridWidget(
            label='Datagrid',
            description='Datagrid',
            rows=2,
        )),
    ))


class ArchAddOnExample(atapi.BaseContent):
    """Example of all of our field types."""

    schema = Schema
    content_icon = 'document_icon.gif'
    archetypes_name = 'ArchAddOn Example'

atapi.registerType(ArchAddOnExample, PROJECTNAME)
